import pickle

import numpy as np
import torch
from numpy import array
from torch import tensor
from matplotlib import cm
from matplotlib import pyplot as plt

from vqr.api import VectorQuantileRegressor
from experiments.utils.metrics import kde, kde_l1, w2_keops
from vqr.solvers.regularized_lse import (
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)
from experiments.datasets.cond_banana import ConditionalBananaDataProvider


def plot_kde(kde_map_1, kde_map_2, l1_distance: float, filename: str):
    fig, axes = plt.subplots(1, 2, figsize=(10, 5))
    axes[0].imshow(
        kde_map_1,
        interpolation="bilinear",
        origin="lower",
        cmap=cm.RdPu,
        extent=(0, 1, 0, 1),
    )
    axes[1].imshow(
        kde_map_2,
        interpolation="bilinear",
        origin="lower",
        cmap=cm.RdPu,
        extent=(0, 1, 0, 1),
    )
    plt.title(f"L1 distance: {l1_distance}")
    plt.savefig(f"{filename}.png")


n = 20000
d = 2
k = 1
T = 50
num_epochs = 20000
linear = False
sigma = 0.1
GPU_DEVICE_NUM = 1
device = f"cuda:{GPU_DEVICE_NUM}" if GPU_DEVICE_NUM is not None else "cpu"
dtype = torch.float32
epsilon = 5e-3

data_provider = ConditionalBananaDataProvider(k=k, d=d, nonlinear=True)
X, Y = data_provider.sample(n=n)

if linear:
    solver = RegularizedDualVQRSolver(
        verbose=True,
        num_epochs=num_epochs,
        epsilon=epsilon,
        lr=2.9,
        gpu=True,
        full_precision=False,
        device_num=GPU_DEVICE_NUM,
        batchsize_y=None,
        batchsize_u=None,
        inference_batch_size=100,
        lr_factor=0.9,
        lr_patience=500,
        lr_threshold=0.5 * 0.01,
    )
else:
    solver = MLPRegularizedDualVQRSolver(
        verbose=True,
        num_epochs=num_epochs,
        epsilon=epsilon,
        lr=0.4,
        gpu=True,
        skip=False,
        batchnorm=False,
        hidden_layers=(2, 10, 20),
        device_num=GPU_DEVICE_NUM,
        batchsize_y=None,
        batchsize_u=None,
        inference_batch_size=100,
        lr_factor=0.9,
        lr_patience=300,
        lr_threshold=0.5 * 0.01,
    )

vqr_est = VectorQuantileRegressor(n_levels=T, solver=solver)

vqr_est.fit(X, Y)

# Generate conditional distributions for the below X's
Xs = [tensor(array([[x] * k]), dtype=dtype) for x in np.linspace(1.0, 3.0, 20)]
kde_l1_dists = []

for cond_X in Xs:
    _, cond_Y_gt = data_provider.sample(n=n, x=cond_X.numpy())
    cond_Y_gt = tensor(cond_Y_gt, dtype=dtype)

    cond_Y_est = vqr_est.sample(n=n, x=cond_X.numpy())
    cond_Y_est = tensor(cond_Y_est, dtype=dtype)

    # w2 distance
    w2_dist = w2_keops(cond_Y_gt, cond_Y_est, device=device)

    # Estimate KDEs
    kde_orig = kde(
        cond_Y_gt,
        grid_resolution=T * 2,
        device=device,
        sigma=sigma,
    )

    kde_est = kde(
        cond_Y_est,
        grid_resolution=T * 2,
        device=device,
        sigma=sigma,
    )

    # Calculate KDE-L1 distance
    kde_l1_dist = kde_l1(
        cond_Y_gt, cond_Y_est, grid_resolution=T * 2, device=device, sigma=sigma
    )
    kde_l1_dists.append(kde_l1_dist)

    # Plot KDEs
    plot_kde(
        kde_orig.T,
        kde_est.T,
        kde_l1_dist,
        f"Y_given_X={cond_X.squeeze().item():.1f}_{linear=}",
    )


with open(f"./kde-l1-dists-{linear=}.pkl", "wb") as f:
    pickle.dump(kde_l1_dists, f)
    print(np.mean(kde_l1_dists))
    f.close()



from __future__ import annotations

from typing import Union, Optional

import cvxpy as cp
import numpy as np
from numpy import ndarray as Array
from scipy.spatial.distance import cdist

from vqr.cvqf import DiscreteVQF, DiscreteCVQF, vector_quantile_levels
from vqr.utils import get_kwargs
from vqr.solvers.base import VQRDiscreteSolver

SIMILARITY_FN_INNER_PROD = lambda x, y: np.dot(x, y)


class CVXVQRSolver(VQRDiscreteSolver):
    """
    Solves the Optimal Transport formulation of Vector Quantile Regression using
    CVXPY as a solver backend.
    """

    def __init__(self, T: int = 50, verbose: bool = False, **cvxpy_kwargs):
        """

        :param T: Number of quantile levels to estimate along each of the d
        dimensions. The quantile level will be spaced uniformly between 0 and 1.
        :param verbose: Whether to be verbose.
        :param cvxpy_kwargs: Any kwargs supported by CVXPY's Problem.solve().
        """
        super().__init__()
        self.T = T
        self.verbose = verbose
        cvxpy_kwargs["verbose"] = verbose
        self.cvxpy_kwargs = cvxpy_kwargs
        self._solver_opts = get_kwargs()

    @classmethod
    def solver_name(cls) -> str:
        return "cvx_primal"

    @property
    def solver_opts(self) -> dict:
        return self._solver_opts

    @property
    def levels_per_dim(self) -> int:
        return self.T

    def solve_vqe(self, Y: Array) -> DiscreteVQF:
        return self._solve(Y)

    def solve_vqr(self, Y: Array, X: Array) -> DiscreteCVQF:
        return self._solve(Y, X)

    def _solve(
        self, Y: Array, X: Optional[Array] = None
    ) -> Union[DiscreteVQF, DiscreteCVQF]:
        T = self.T
        N = len(Y)
        Y = np.reshape(Y, (N, -1))

        ones = np.ones(shape=(N, 1))
        if X is None:
            X = ones
        else:
            X = np.reshape(X, (N, -1))
            X = np.concatenate([ones, X], axis=1)

        k: int = X.shape[1] - 1  # Number of features (can be zero)
        d: int = Y.shape[1]  # number or target dimensions

        X_bar = np.mean(X, axis=0, keepdims=True)  # (1, k+1)

        # All quantile levels
        Td: int = T**d
        U: Array = vector_quantile_levels(T, d)
        assert U.shape == (Td, d)

        # Pairwise distances (similarity)
        S: Array = cdist(U, Y, SIMILARITY_FN_INNER_PROD)  # (Td, d) and (N, d)

        # Optimization problem definition: optimal transport formulation
        one_N = np.ones([N, 1])
        one_T = np.ones([Td, 1])
        Pi = cp.Variable(shape=(Td, N))
        Pi_S = cp.sum(cp.multiply(Pi, S))
        constraints = [
            Pi @ X == 1 / Td * one_T @ X_bar,
            Pi >= 0,
            one_T.T @ Pi == 1 / N * one_N.T,
        ]
        problem = cp.Problem(objective=cp.Maximize(Pi_S), constraints=constraints)

        # Solve the problem
        problem.solve(**self.cvxpy_kwargs)

        # Obtain the lagrange multipliers Alpha (A) and Beta (B)
        AB: Array = constraints[0].dual_value
        AB = np.reshape(AB, newshape=[Td, k + 1])
        A = AB[:, [0]]  # A is (T**d, 1)
        if k == 0:
            return DiscreteVQF(T, d, U, A, refine=False)
        else:
            B = AB[:, 1:]  # B is (T**d, k)
            return DiscreteCVQF(T=T, d=d, U=U, A=A, B=B, k_in=k)



from __future__ import annotations

import warnings
from typing import Any, Dict, List, Tuple, Callable, Optional, Sequence
from itertools import repeat

import ot
import numpy as np
from numpy import ndarray as Array
from sklearn.utils import check_array


class VQFBase:
    """
    Common functionality for VQF/CVQF.
    """

    def __init__(
        self,
        d: int,
        solution_metrics: Optional[Dict[str, Any]] = None,
    ):
        """
        :param d: Dimension of target variable.
        :param solution_metrics: Optional key-value pairs which can contain any
        metric values which are tracked by the solver, such as losses, runtimes, etc.
        The keys in this dict are solver-specific and should be specified in the
        documentation of the corresponding solver.
        """
        assert d >= 1
        assert solution_metrics is None or isinstance(solution_metrics, dict)
        self._d = d
        self._solution_metrics = solution_metrics or {}

    @property
    def dim_y(self) -> int:
        """
        :return: The dimension d, of the target variable (Y).
        """
        return self._d

    @property
    def metrics(self):
        """
        :return: A dict containing solver-specific metrics about the solution.
        """
        return self._solution_metrics.copy()


class DiscreteVQFBase(VQFBase):
    """
    Common functionality for DiscreteVQF/DiscreteCVQF.
    """

    def __init__(
        self,
        d: int,
        T: int,
        U: Array,
    ):
        """
        :param T: Number of vector quantile levels per dimension.
        :param U: Array of shape (T**d, d). Contains the d-dimensional grid on
        which the vector quantiles are defined. If can be decoded back into a
        meshgrid-style sequence of arrays using :obj:`decode_quantile_grid`
        """
        assert U.shape == (T**d, d)
        self._T = T
        self._U = U

    @property
    def levels_per_dim(self) -> int:
        """
        :return: T, the number of quantile levels to estimate along each of the d
        dimensions. The quantile level will be spaced uniformly between 0 and 1.
        """
        return self._T

    @property
    def quantile_grid(self) -> Array:
        """
        :return: All the discrete quantile levels of this quantile function.
        A (d+1)-dimensional array of shape (d, T, T, ... T), where the first axis
        indexes different quantile surfaces.
        """
        T, d = self._T, self._d
        U = _decode_quantile_grid(T, d, self._U)
        U = np.stack(U, axis=0)  # Stack into shape (d,T,T,...,T)
        assert U.shape == tuple([d, *[T] * d])
        return U

    @property
    def quantile_levels(self) -> Array:
        """
        :return: An array containing the levels at which the vector quantiles were
            estimated along each target dimension.
        """
        return quantile_levels(self._T)


class VQF(VQFBase):
    """
    Represents a vector quantile function, Q_{Y}(u) or a conditional vector quantile
    function conditioned on a specific value of X=x, i.e. Q_{Y|X=x}(u).

    Y is assumed to be a d-dimensional variable (d>=1).
    """

    def __init__(
        self,
        d: int,
        solution_metrics: Optional[Dict[str, Any]] = None,
    ):
        """
        :param d: Dimension of target variable.
        :param solution_metrics: Optional key-value pairs which can contain any
        metric values which are tracked by the solver, such as losses, runtimes, etc.
        The keys in this dict are solver-specific and should be specified in the
        documentation of the corresponding solver.
        """
        VQFBase.__init__(self, d=d, solution_metrics=solution_metrics)

    def evaluate(self, u: Array) -> Array:
        """
        Evaluates the VQF at quantile level u.

        :param u: d-dimensional vector quantile level of shape (d,). Each value
        should be in [0, 1].
        :return: A d-dimensional vector quantile of shape (d,).
        """
        raise NotImplementedError(f"Continuous VQF not yet implemented")

    __call__ = evaluate


class CVQF(VQF):
    """
    Represents a conditional vector quantile function, Q_{Y|X}(u;x).

    Y is assumed to be a d-dimensional variable (d>=1), and X a k-dimensional variable.
    """

    def __init__(
        self,
        d: int,
        X_transform: Optional[Callable[[Array], Array]] = None,
        k_in: Optional[int] = None,
        solution_metrics: Optional[Dict[str, Any]] = None,
    ):
        """
        :param d: Dimension of target variable.
        :param X_transform: Transformation to apply to covariates (X) for non-linear
        VQR. Must be provided together with k_in. The transformation is assumed to
        take input of shape (N, k_in) and return output of shape (N, k).
        :param k_in: Covariates input dimension of the X_transform.
        If X_transform is None, must be None or zero.
        :param solution_metrics: Optional key-value pairs which can contain any
        metric values which are tracked by the solver, such as losses, runtimes, etc.
        The keys in this dict are solver-specific and should be specified in the
        documentation of the corresponding solver.
        """
        super().__init__(d=d, solution_metrics=solution_metrics)
        assert k_in >= 1
        self._k_in = k_in
        self._X_transform = X_transform

    @property
    def dim_x(self) -> int:
        """
        :return: The dimension k, of the covariates (X) which are expected to be
        passed in to obtain conditional quantiles.
        """
        # If there was an X_transform, the input dimension of that is what we expect to
        # receive.
        return self._k_in

    def evaluate(self, u: Array, x: Array = None) -> Array:
        """
        Evaluates the CVQF at quantile level u for covariates x.

        :param u: d-dimensional vector quantile level of shape (d,). Each value
        should be in [0, 1].
        :param x: k-dimensional covariates (features) vector, of shape (k,) or (1,k).
        Note: None is only supported to comply with VQF API; x=None will be
        treated as x=0 which may not be meaningful for the data distribution.
        :return: A d-dimensional vector quantile of shape (d,).
        """
        raise NotImplementedError(f"Continuous CVQF not yet implemented")

    def condition(self, x: Array) -> VQF:
        """
        Conditions the CVQF on a specific covariate vector, x.

        :param x: The covariates on which to condition of shape (k,) or (1,k).
        :return: A quantile function which then only depends on the level u of shape
        (d,).
        """
        raise NotImplementedError(f"Continuous CVQF not yet implemented")


class DiscreteCVQF(CVQF, DiscreteVQFBase):
    """
    Represents a discrete conditional vector quantile function, Q_{Y|X}(u;x) where:

    - Y is a d-dimensional target variable (d>=1).
    - X is a k-dimensional covariates variable (k>=1).
    - u is a d-dimensional quantile level, discretized to T levels per dimension.
    - Q_{Y|X=x}(u) is the d-dimensional quantile of Y|X=x at level u.

    Should be constructed by :class:`VQRSolver`s, not manually.
    """

    def __init__(
        self,
        T: int,
        d: int,
        U: Array,
        A: Array,
        B: Array,
        k_in: int,
        X_transform: Optional[Callable[[Array], Array]] = None,
        solution_metrics: Optional[Dict[str, Any]] = None,
    ):
        """
        :param d: Dimension of target variable.
        :param U: Array of shape (T**d, d). Contains the d-dimensional grid on
        which the vector quantiles are defined. If can be decoded back into a
        meshgrid-style sequence of arrays using :obj:`decode_quantile_grid`
        :param A: Array of shape (T**d, 1). Contains the  the regression intercept
        variable. This can be decoded into the d vector quantiles of Y using the
        :obj:`decode_quantile_values` function.
        :param B: Array of shape (T**d, k). Contains the  regression coefficients.
        Will be None if the input was an estimation problem (X=None) instead of a
        regression problem. The k is the dimension of the covariates (X). If an
        X_transform is provided, k corresponds to the dimension AFTER the
        transformation.
        :param X_transform: Transformation to apply to covariates (X) for non-linear
        VQR. Must be provided together with k_in. The transformation is assumed to
        take input of shape (N, k_in) and return output of shape (N, k).
        :param k_in: Covariates input dimension of the X_transform.
        If X_transform is None, must be None or zero.
        :param solution_metrics: Optional key-value pairs which can contain any
        metric values which are tracked by the solver, such as losses, runtimes, etc.
        The keys in this dict are solver-specific and should be specified in the
        documentation of the corresponding solver.
        """
        CVQF.__init__(
            self,
            d=d,
            k_in=k_in,
            X_transform=X_transform,
            solution_metrics=solution_metrics,
        )
        DiscreteVQFBase.__init__(self, d=d, T=T, U=U)

        # Validate dimensions
        assert all(x is not None for x in [T, d, U, A])
        assert U.ndim == 2 and A.ndim == 2
        assert U.shape[0] == A.shape[0]
        assert U.shape[0] == T**d
        assert U.shape[1] == d
        assert A.shape[1] == 1
        assert B.ndim == 2 and B.shape[0] == T**d
        assert k_in

        self._A = A
        self._B = B
        self._k = B.shape[1]

    def evaluate(self, u: Array, x: Array = None, refine: bool = False) -> Array:
        """
        Evaluates the CVQF at quantile level u for covariates x.

        :param u: d-dimensional vector quantile level of shape (d,). Each value
        should be in [0, 1].
        :param x: k-dimensional covariates (features) vector, of shape (k,) or (1,k).
        Note: None is only supported to comply with VQF API; x=None will be
        treated as x=0 which may not be meaningful for the data distribution.
        :param refine: Whether to refine the conditional quantile function using vector
        monotone rearrangement.
        :return: A d-dimensional vector quantile of shape (d,).
        """
        if x is None:
            # This is only supported to conform to VQF.evaluate(u).
            warnings.warn(f"Evaluating DiscreteCVQF with X=None not recommended")
            x = np.zeros(shape=(1, self._k_in))

        return self.condition(x, refine=refine).evaluate(u)

    def condition(self, x: Array, refine: bool = False) -> DiscreteVQF:
        """
        Conditions the CVQF on a specific covariate vector, x.

        :param x: The covariates on which to condition, shape (1,k) or (k,).
        :param refine: Whether to refine the conditional quantile function using vector
        monotone rearrangement.
        :return: A DiscreteVQF instance corresponding to the given covariates X.
        """

        x = x.reshape(1, -1)
        check_array(x, ensure_2d=True, allow_nd=False)

        z = x  # z represents the transformed x
        if self._X_transform is not None:
            _, k_in = x.shape
            if k_in != self._k_in:
                raise ValueError(
                    f"VQR model was trained with X_transform expecting k_in"
                    f"={self._k_in}, but got covariates with {k_in=} features."
                )

            z = self._X_transform(x)

        _, k = z.shape
        if k != self._k:
            raise ValueError(
                f"VQR model was fitted with k={self._k}, "
                f"but got data with {k=} features."
            )

        B = self._B  # (T**d, k)
        A = self._A  # (T**d, 1)
        Y_hat = B @ z.T + A  # result is (T**d, 1)
        Y_hat = Y_hat  # (1, T**d)

        return DiscreteVQF(
            T=self._T,
            d=self._d,
            U=self._U,
            A=Y_hat,
            x=x,
            solution_metrics=self._solution_metrics,
            refine=refine,
        )

    @property
    def dim_x(self) -> int:
        # If there was an X_transform, the input dimension of that is what we expect to
        # receive.
        return self._k_in or self._k


class DiscreteVQF(VQF, DiscreteVQFBase):
    """
    Represents a discretized vector-valued quantile function Q_{Y|X=x}(u)
    of the variable Y|X=x (i.e. pre-conditioned on a specific x), or unconditioned
    vector quantile function Q_{Y}(u), where:

    - Y is a d-dimensional target variable.
    - X is a k-dimensional covariates variable with specific realization x.
    - u is a d-dimensional quantile level, discretized to T levels per dimension.
    - Q_{Y|X=x}(u) is the d-dimensional quantile of Y|X=x at level u.

    Should be constructed by :class:`VQESolver`s, not manually.

    Instances of this class expose both the d quantile surfaces of the function,
    and also allow evaluating the function at a given quantile level u.

    - Iterating over an instance or indexing it yields the d quantile surfaces,
      e.g. list(q) or q[i].
    - Using an instance as a function evaluates the quantile function, eg. q(u).
    """

    def __init__(
        self,
        T: int,
        d: int,
        U: Array,
        A: Array,
        x: Optional[Array] = None,
        solution_metrics: Optional[Dict[str, Any]] = None,
        refine: bool = False,
    ):
        """
        :param U: Array of shape (T**d, d). Contains the d-dimensional grid on
        which the vector quantiles are defined. If can be decoded back into a
        meshgrid-style sequence of arrays using :obj:`decode_quantile_grid`
        :param A: Array of shape (T**d, 1). Contains the  the regression intercept
        variable. This can be decoded into the d vector quantiles of Y using the
        :obj:`decode_quantile_values` function.
        :param solution_metrics: Optional key-value pairs which can contain any
        metric values which are tracked by the solver, such as losses, runtimes, etc.
        The keys in this dict are solver-specific and should be specified in the
        documentation of the corresponding solver.
        :param x: The covariates vector that was used to condition a CVQF to produce
        this VQF, if any.
        :param refine: Whether to refine the conditional quantile function using vector
        monotone rearrangement.
        """
        VQF.__init__(self, d=d, solution_metrics=solution_metrics)
        DiscreteVQFBase.__init__(self, d=d, T=T, U=U)

        # Validate dimensions
        assert all(x is not None for x in [T, d, U, A])
        assert U.ndim == 2 and A.ndim == 2
        assert U.shape[0] == A.shape[0]
        assert U.shape[0] == T**d
        assert U.shape[1] == d
        assert A.shape[1] == 1

        self._x = x
        self._k = 0 if x is None else x.shape[1]
        self._A = A

        Q = _decode_quantile_values(self._T, self._d, A)
        if refine:
            Q = vector_monotone_rearrangement(Q)
        self._is_refined = refine

        self._Q = np.stack(Q, axis=0)  # Stack into shape (d,T,T,...,T)
        assert self._Q.shape == tuple([d, *[T] * d])

    def refine(self) -> DiscreteVQF:
        """
        :return: A version of this VQF, refined with vector monotone rearrangeament (
        VMR). Refinement guarantees that the VQF will be a co-monotonic with u.
        """
        if self.is_refined:
            return self
        return DiscreteVQF(
            self._T,
            self._d,
            self._U,
            self._A,
            self._x,
            self._solution_metrics,
            refine=True,
        )

    @property
    def is_refined(self) -> bool:
        return self._is_refined

    @property
    def values(self) -> Array:
        """
        :return: All the discrete values of this quantile function.
        A (d+1)-dimensional array of shape (d, T, T, ... T), where the first axis
        indexes different quantile surfaces.
        """
        return self._Q

    def __iter__(self):
        """
        :return: An iterator over the quantile surfaces (Qs) of this quantile function.
        Yields d elements, each a d-dimensional array of shape (T, T, ..., T).
        """
        for d_idx in range(self._d):
            yield self._Q[d_idx]

    def __len__(self):
        """
        :return: Number of dimensions of this quantile function.
        """
        return self._d

    def __getitem__(self, d_idx: int) -> Array:
        """
        :param d_idx: An index of a dimension of the target variable, in range [0, d).
        :return: Quantile surface for that dimension.
        """
        return self._Q[d_idx]

    def evaluate(self, u: Array) -> Array:
        if not np.ndim(u) == 1 or not len(u) == self._d:
            raise ValueError(f"u must be of shape (d,), got shape {u.shape}")

        if not np.all((u >= 0) & (u <= 1)):
            raise ValueError(
                f"u must contain indices of quantile levels, each in [0, 1]"
            )

        u_idx = np.floor(u * (self._T - 1)).astype(np.int)
        q_idx = (slice(None), *u_idx)
        q = self._Q[q_idx]
        assert q.shape == (self._d,)
        return q


def quantile_levels(T: int) -> Array:
    """
    Creates a vector of evenly-spaced quantile levels between zero and one.
    :param T: Number of levels to create.
    :return: An array of shape (T,).
    """
    return (np.arange(T) + 1) * (1 / T)


def vector_quantile_levels(T: int, d: int) -> Array:
    """
    Constructs an array of vector-quantile levels. Each level is a d-dimensional
    vector with entries in [0,1].

    The quantile levels are returned stacked along the first axis of the result.

    :param T: Number of levels in each dimension.
    :param d: Dimension of target variable.
    :return: A (T^d, d) array with a different level at each "row".
    """
    # Number of all quantile levels
    Td: int = T**d

    # Quantile levels grid: list of grid coordinate matrices, one per dimension
    # d arrays of shape (T,..., T)
    U_grids: Sequence[Array] = np.meshgrid(*([quantile_levels(T)] * d))

    # Stack all nd-grid coordinates into one long matrix, of shape (T**d, d)
    U: Array = np.stack([U_grid.reshape(-1) for U_grid in U_grids], axis=1)
    assert U.shape == (Td, d)

    return U


def _decode_quantile_values(T: int, d: int, Y_hat: Array) -> Sequence[Array]:
    """
    Decodes the regression coefficients of a VQR solution into vector quantile values.
    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
        problem.
    :param Y_hat: The regression output, of shape (T**d, 1).
    :return: A sequence of length d of vector quantile values. Each element j in the
        sequence is a d-dimensional array of shape (T, T, ..., T) containing the
        vector quantiles values of j-th variable in Y, i.e., the quantiles of Y_j|Y_{-j}
        where Y_{-j} means all the variables in Y except the j-th.
    """
    Q = np.reshape(Y_hat, newshape=(T,) * d)

    Q_functions: List[Array] = [np.array([np.nan])] * d
    for axis in reversed(range(d)):
        # Calculate derivative along this axis
        dQ_du = (1 / T) * np.diff(Q, axis=axis)

        # Duplicate first "row" along axis and insert it first
        pad_with = [
            (0, 0),
        ] * d
        pad_with[axis] = (1, 0)
        dQ_du = np.pad(dQ_du, pad_width=pad_with, mode="edge")

        Q_functions[d - 1 - axis] = dQ_du * T**2

    return tuple(Q_functions)


def _decode_quantile_grid(T: int, d: int, U: Array) -> Sequence[Array]:
    """
    Decodes stacked vector quantile levels (the output of the vector_quantile_level()
    function) into a "meshgrid" of the evaluation points of the vector quantile
    function.

    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
    problem.
    :param U: The encoded grid U, of shape (T**d, d).
    :return: A sequence length d. Each array in the sequence is a d-dimensional
    array of shape (T, T, ..., T), which together represent the d-dimensional grid
    on which the vector quantiles were evaluated.
    """
    return tuple(np.reshape(U[:, dim], newshape=(T,) * d) for dim in range(d))


def get_d_T(Qs: Sequence[Array]) -> Tuple[int, int]:
    """
    Helper for obtaining d and T parameters from VQF surfaces.

    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :return: d and T. Will raise if shapes are not consistent with the above.
    """
    shapes = tuple(Q.shape for Q in Qs)

    d = len(shapes)
    T = shapes[0][0]

    # Validate
    expected_shape = (T,) * d
    if not all(shape == expected_shape for shape in shapes):
        raise ValueError(f"Quantile surfaces have unexpected shapes: {shapes}")

    return d, T


def inversion_sampling(n: int, Qs: Sequence[Array]):
    """
    Generates samples from the variable Y based on it's fitted
    quantile function, using inversion-transform sampling.

    :param n: Number of samples to generate.
    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :return: Samples obtained from this quantile function, of shape (n, d).
    """

    d, T = get_d_T(Qs)

    # Samples of points on the quantile-level grid
    Us = np.random.randint(0, T, size=(n, d))

    # Sample from Y|X=x
    Y_samp = np.empty(shape=(n, d))
    for i, U_i in enumerate(Us):
        # U_i is a vector-quantile level, of shape (d,)
        Y_samp[i, :] = np.array([Q_d[tuple(U_i)] for Q_d in Qs])

    return Y_samp


def quantile_contour(Qs: Sequence[Array], alpha: float = 0.05) -> Tuple[Array, Array]:
    """
    Creates a contour of points in d-dimensional space which surround the region in
    which 100*(1-2*alpha)^d percent of the distribution (that was corresponds to a
    given quantile function) is contained.

    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param alpha: Confidence level for the contour.
    :return: A tuple of two arrays, each of shape (n, d).
    The first contains the points along the d-dimensional
    contour, and the second contains the d-dimensional index of each point.
    """

    if not 0 < alpha < 0.5:
        raise ValueError(f"Got {alpha=}, but must be in (0, 0.5)")

    d, T = get_d_T(Qs)
    lo = int(np.round(T * alpha))
    hi = int(min(np.round(T * (1 - alpha)), T - 1))

    # Will contain d lists of points, each element will be (d,)
    contour_points_list = [[] for _ in range(d)]
    contour_index_list = [[] for _ in range(d)]

    for i, Q in enumerate(Qs):
        for j in range(d):
            for lo_hi, _ in zip([lo, hi], range(d)):
                # The zip here is just to prevent a loop when d==1.
                # This keeps all dims fixed on either lo or hi, while the j-th
                # dimension is sweeped from lo to hi.
                # idx: List[Union[int, slice]] = [lo_hi] * d
                # idx[j] = slice(lo, hi)

                idxs = [repeat(lo_hi)] * d
                idxs[j] = range(lo, hi)
                zipped_idxs = tuple(zip(*idxs))  # list of tuples e.g. [(u1, u2), ...]
                # d lists, e.g. [(u1, u1, ...), (u2, u2, ...)]
                grouped_indices = tuple(zip(*zipped_idxs))

                contour_points_list[i].extend(Q[grouped_indices])
                contour_index_list[i].extend(zipped_idxs)

    # Contours generated by accessing the same indices in each quantile surface
    assert all(
        np.allclose(contour_index_list[0], contour_index_list[i]) for i in range(1, d)
    )
    contour_points = np.array(contour_points_list).T  # (N, d)
    contour_indices = np.array(contour_index_list[0])  # (N, d)
    return contour_points, contour_indices


def vector_monotone_rearrangement(
    Qs: Sequence[Array], max_iters: int = 2e6
) -> Sequence[Array]:
    """
    Performs vector monotone rearrangement. Can be interpreted as "vector sorting".

    A vector-extension for the quantile rearrangement idea proposed in

    Victor Chernozhukov, Iván Fernández‐Val, Alfred Galichon.
    Quantile and probability curves without crossing.
    Econometrica, 2010.

    Solves an exact OT problem using POT's emd solver with U @ Y.T as the cost matrix.
    Rearrangement is performed by multiplication with the permutation matrix Pi.

    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param max_iters: Maximum number of iterations for the emd solver.
    :return: Rearranged quantile surfaces per dimension of Y. A sequence of length d
    where each element is of shape (T, T, ..., T).
    """
    d, T = get_d_T(Qs)
    U: Array = vector_quantile_levels(T, d)
    Y: Array = np.stack([Q.ravel() for Q in Qs], axis=-1)
    pi = ot.emd(M=-U @ Y.T, a=[], b=[], numItermax=max_iters)
    rearranged_Y = (T**d) * pi @ Y
    rearranged_Qs = [rearranged_Y[:, i].reshape((T,) * d) for i in range(d)]
    return rearranged_Qs


def check_comonotonicity(Qs: Sequence[Array], Us: Sequence[Array]) -> Array:
    """
    Measures co-monotonicity defined as (u_i - u_j).T @ (Q(u_i) - Q(u_j)) for all i, j.
    Results in a T^d x T^d symmetric matrix.  If co-monotonicity is satisfied,
    all entries in this matrix are positive. Negative entries in this matrix
    represent the quantile crossing problem (and its analogue in higher dimensions).

    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param Us: Quantile levels per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :return: A T^d x T^d symmetric matrix that measures co-monotonicity between all
    pairs of quantile levels and quantile values. The (i, j)^th entry in the matrix
    measures co-monotonicity between quantile levels  u_i and u_j and the quantile
    values Q(u_i) and Q(u_j).
    """
    d, T = get_d_T(Qs)
    levels = np.stack([level.ravel() for level in Us])
    quantiles = np.stack([Q.ravel() for Q in Qs])

    pairwise_diff_levels = (
        levels.reshape(d, 1, T**d) - levels.reshape(d, T**d, 1)
    ).reshape(d, T ** (2 * d))

    pairwise_diff_quantiles = (
        quantiles.reshape(d, 1, T**d) - quantiles.reshape(d, T**d, 1)
    ).reshape(d, T ** (2 * d))

    all_pairs_inner_prod = np.sum(
        pairwise_diff_levels * pairwise_diff_quantiles, axis=0
    ).reshape(T**d, T**d)

    return all_pairs_inner_prod



import os
import shutil
import itertools as it
from typing import Tuple, Sequence
from pathlib import Path

import numpy as np
import pytest
from numpy.linalg import norm
from _pytest.fixtures import FixtureRequest

from tests import TESTS_OUT_DIR
from vqr.cvqf import check_comonotonicity, vector_quantile_levels
from experiments.log import setup_logging

setup_logging()


@pytest.fixture(scope="function")
def test_out_dir(request: FixtureRequest) -> Path:
    """
    Creates a unique output directory in which a test can write some results.
    It will be cleared if it existed before.

    The directory will live under :obj:`TESTS_OUT_DIR` and be named according to the
    test class and test method.
    :return: A :class:`Path` representing the directory.
    """
    return _test_out_dir(request)


@pytest.fixture(scope="class")
def test_out_dir_class(request: FixtureRequest) -> Path:
    """
    Same as tests_out_dir but for class-level fixtures.

    Note: the returned dir is not cleared!
    For class-level dirs, we can't clear because multiple processes running tests
    in the same class will remove each-others folders.
    """
    return _test_out_dir(request, clear=False)


@pytest.fixture(scope="class")
def test_out_dir_class_pid(request: FixtureRequest) -> Path:
    """
    Same as tests_out_dir but for class-level fixtures.
    A separate out dir will be created for every test process.
    It will be cleared if it existed before.
    """
    return _test_out_dir(request, clear=True, with_pid=True)


def _test_out_dir(request: FixtureRequest, clear: bool = True, with_pid: bool = False):
    filename, *classname_testname = str.split(request.node.nodeid, "::")
    out_dir = TESTS_OUT_DIR.joinpath(*classname_testname)
    if with_pid:
        out_dir = out_dir.joinpath(f"pid_{os.getpid()}")

    # Clear any previous outputs to prevent any possibility of previous results to
    # affect the current run.
    if clear and out_dir.is_dir():
        shutil.rmtree(out_dir)

    os.makedirs(out_dir, exist_ok=True)
    return out_dir


def _test_monotonicity(
    Us: Sequence[np.ndarray],
    Qs: Sequence[np.ndarray],
    projection_tolerance: float = 0.0,
    offending_proportion_limit: float = 0.005,
):
    offending_projections, projections = monotonicity_offending_projections(
        Qs, Us, projection_tolerance
    )
    n_c, n = len(offending_projections), len(projections)

    offending_proportion = n_c / n
    if offending_projections:
        q = [0.01, 0.1, 0.25, 0.5, 0.75, 0.9, 0.99]
        print(f"err quantiles: {np.quantile(offending_projections, q=q)}")
        print(f"all quantiles: {np.quantile(projections, q=q)}")
        print(f"{n=}, {n_c=}, {n_c/n=}")

    assert offending_proportion < offending_proportion_limit


def monotonicity_offending_projections(
    Qs: Sequence[np.ndarray],
    Us: Sequence[np.ndarray],
    projection_tolerance: float,
) -> Tuple[Sequence[float], Sequence[float]]:
    assert len(Qs) == len(Us)

    pairwise_comonotonicity_mat = check_comonotonicity(Qs=Qs, Us=Us)
    offending_projections = pairwise_comonotonicity_mat[
        np.where(np.triu(pairwise_comonotonicity_mat) < projection_tolerance)
    ].tolist()
    projections = np.triu(pairwise_comonotonicity_mat).ravel().tolist()

    return offending_projections, projections



from vqr.cvqf import DiscreteCVQF
from vqr.solvers.cvx import CVXVQRSolver
from vqr.solvers.pot import POTVQESolver
from vqr.solvers.base import Solver, VQESolver, VQRSolver, DiscreteSolver
from vqr.solvers.regularized_lse import (
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)



from typing import Union, Optional

import torch
from numpy import ndarray
from torch import Tensor


def ensure_numpy(x: Union[ndarray, Tensor]) -> ndarray:
    """
    Ensures that the input tensor is a numpy array, and converts it if not.
    :param x: Either a numpy array or a torch tensor.
    :return: The input as a numpy array.
    """
    if isinstance(x, Tensor):
        return x.clone().detach().numpy()
    elif isinstance(x, ndarray):
        return x
    else:
        raise ValueError(f"Unexpected type, {type(x)=}")


def ensure_torch(
    x: Union[ndarray, Tensor],
    dtype: Optional[torch.dtype] = None,
    device: Optional[str] = None,
) -> Tensor:
    """
    Ensures that the input tensor is a torch tensor of a specified type on a
    specified device, and converts it if not.
    :param x: Either a numpy array or a torch tensor.
    :param dtype: The datatype of the tensor to return. Will be converted if necessary.
    :param device: The device on which the result should be. Will be copied if
    necessary.
    :return: The input as a torch tensor with the specified type and device.
    """
    if isinstance(x, Tensor):
        return x.to(dtype=dtype, device=device)
    elif isinstance(x, ndarray):
        return torch.from_numpy(x).to(dtype=dtype, device=device)
    else:
        raise ValueError(f"Unexpected type, {type(x)=}")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Type, TypeVar, Optional

from numpy import ndarray as Array

from vqr.cvqf import VQF, CVQF, DiscreteCVQF

TSolver = TypeVar("TSolver", bound="Solver")


class Solver(ABC):
    @classmethod
    @abstractmethod
    def solver_name(cls) -> str:
        """
        :return: An identifier for this solver.
        """
        pass

    @property
    @abstractmethod
    def solver_opts(self) -> dict:
        """
        :return: Implementation-specific options use to configure this solver.
        """
        pass

    def copy(self: Type[TSolver], **solver_opts) -> TSolver:
        """
        Creates a copy of this solver, optionally with some parameters overridden.

        :param solver_opts: Solver options which will override the existing options (
        can be a subset).
        :return: A new solver of the same type as self, initialized with the current
        options and given overrides.
        """
        new_opts = {**self.solver_opts, **solver_opts}
        return type(self)(**new_opts)


class VQESolver(Solver, ABC):
    """
    Abstraction of a method for solving the Vector Quantile Estimation (VQE) problem.
    """

    @abstractmethod
    def solve_vqe(self, Y: Array) -> VQF:
        """
        Solves the VQE problem in an implementation-specific way.

        :param Y: The regression target variable, of shape (N, d) where N is the
        number of samples and d is the dimension of the target, which is also the
        dimension of the quantiles which will be estimated.
        :return: A VQF representing Q_{Y}.
        """
        pass


class VQRSolver(VQESolver, ABC):
    """
    Abstraction of a method for solving the Vector Quantile Regression (VQR) problem.
    """

    @abstractmethod
    def solve_vqr(
        self,
        Y: Array,
        X: Array,
    ) -> CVQF:
        """
        Solves the provided VQR problem in an implementation-specific way.

        :param Y: The regression target variable, of shape (N, d) where N is the
        number of samples and d is the dimension of the target, which is also the
        dimension of the quantiles which will be estimated.
        :param X: The regression input (features, covariates), of shape (N, k),
        where k is the number of features.
        :return: A CVQF representing Q_{Y|X}.
        """
        pass


class DiscreteSolver(Solver, ABC):
    """
    Represents a VQR solver which solves the discrete problem, i.e., obtains an
    estimate of the conditional vector quantile function discretized at T quantile
    levels per dimension.
    """

    @property
    @abstractmethod
    def levels_per_dim(self) -> int:
        """
        :return: T, the number of quantile levels to estimate along each of the d
        dimensions. The quantile level will be spaced uniformly between 0 and 1.
        """
        pass


class VQRDiscreteSolver(VQRSolver, DiscreteSolver, ABC):
    """
    Represents a VQR solver which solves the discrete problem, i.e., obtains an
    estimate of the conditional vector quantile function discretized at T quantile
    levels per dimension.
    """

    pass


class VQEDiscreteSolver(VQESolver, DiscreteSolver, ABC):
    """
    Represents a VQE solver which solves the discrete problem, i.e., obtains an
    estimate of the vector quantile function discretized at T quantile
    levels per dimension.
    """

    pass



import inspect
from typing import Sequence


def get_kwargs(ignore: Sequence[str] = ()) -> dict:
    """
    Assembles the arguments and kwargs of the caller.

    :param ignore: Keys to ignore.
    :return: The args/kwargs of the calling function as a dict.
    """
    frame = inspect.currentframe().f_back

    ignore_set = {"self", *ignore}
    arg_names, _, _, arg_values = inspect.getargvalues(frame)
    kwargs = {
        arg_name: arg_values[arg_name]
        for arg_name in arg_names
        if arg_name not in ignore_set
    }

    return kwargs



from typing import Optional, Sequence

import numpy as np
from numpy import array
from numpy import ndarray as Array
from matplotlib import pyplot as plt

from experiments.datasets.base import DataProvider


class SyntheticGlassesDataProvider(DataProvider):
    @property
    def k(self) -> int:
        return 1

    @property
    def d(self) -> int:
        return 1

    def sample_x(self, n: int) -> Array:
        return np.linspace(0, 1, num=n)

    def sample(self, n: int, x: Optional[array] = None) -> Sequence[array]:
        Z = np.random.rand(n) > 0.5
        Y_i = (
            5 * np.sin(np.linspace(start=0.0, stop=3 * np.pi, num=n))
            + 2.5
            + 2 * np.random.beta(a=0.5, b=1, size=(n,))
        )
        Y_j = (
            5 * np.sin(np.linspace(start=np.pi, stop=4 * np.pi, num=n))
            + 2.5
            - 2 * np.random.beta(a=0.5, b=1, size=(n,))
        )
        return self.sample_x(n), Z * Y_i + (~Z) * Y_j


if __name__ == "__main__":
    dp = SyntheticGlassesDataProvider()
    X_, Y_ = dp.sample(n=3000)
    plt.plot(X_, Y_, ".")
    plt.show()



from typing import Union, Sequence

import torch
from numpy import array
from torch import Tensor, nn, eye, diag
from torch import ones as ones_th


class QuadraticModel(nn.Module):
    def __init__(self, k=2):
        super().__init__()
        self.C1 = nn.Parameter(eye(k, k, dtype=torch.float32, requires_grad=True))
        self.C2 = nn.Parameter(
            ones_th(k, dtype=torch.float32, requires_grad=True),
        )
        self.C3 = nn.Parameter(
            torch.tensor(array([0.0]), dtype=torch.float32, requires_grad=True)
        )
        self.batch_norm = nn.BatchNorm1d(
            num_features=k, affine=False, track_running_stats=True
        )

    def forward(self, X: Tensor):
        return self.batch_norm(diag(X @ self.C1 @ X.T)[:, None] + self.C2 * X + self.C3)


NLS = {
    "relu": torch.nn.ReLU,
    "tanh": nn.Tanh,
    "sigmoid": nn.Sigmoid,
    "softmax": nn.Softmax,
    "logsoftmax": nn.LogSoftmax,
    "silu": nn.SiLU,
    "lrelu": nn.LeakyReLU,
}


class MLP(nn.Module):
    """
    A Simple MLP.

    Structure is:
    FC -> [BN] -> ACT -> [DROPOUT] -> ... FC -> [BN] -> ACT -> [DROPOUT] -> FC

    Note that BatchNorm and Dropout are optional and the MLP always ends with an FC.
    """

    def __init__(
        self,
        in_dim: int,
        hidden_dims: Union[str, Sequence[int]],
        nl: Union[str, nn.Module] = "silu",
        skip: bool = False,
        batchnorm: bool = False,
        dropout: float = 0,
    ):
        """
        :param in_dim: Input feature dimension.
        :param hidden_dims: Hidden dimensions. Will be converted to FC layers. Last
        entry is the output dimension. If a string is provided, it will be parsed as
        a comma-separated list of integer values, e.g. '12,34,46,7'.
        :param nl: Non-linearity to use between FC layers.
        :param skip: Whether to use a skip-connection (over all layers). This
        requires that in_dim==out_dim, thus if skip==True and hidden_dims[-1]!=in_dim
        then the last hidden layer will be changed to produce an output of size in_dim.
        :param batchnorm: Whether to use Batch Normalization
        (before each non-linearity).
        :param dropout: Whether to use dropout (after each non-linearity).
        Zero means no dropout, otherwise means dropout probability.
        """
        super().__init__()

        if isinstance(hidden_dims, str):
            try:
                hidden_dims = tuple(map(int, hidden_dims.strip(", ").split(",")))
            except (TypeError, ValueError) as e:
                raise ValueError(
                    "When hidden_dims is a string it must be a comma-separated "
                    "sequence of integers, e.g. '11,22,34,45'"
                ) from e

        if not hidden_dims:
            raise ValueError(f"got {hidden_dims=} but must have at least one")

        if isinstance(nl, nn.Module):
            non_linearity = nl
        else:
            if nl not in NLS:
                raise ValueError(f"got {nl=} but must be one of {[*NLS.keys()]}")
            non_linearity = NLS[nl]

        if not 0 <= dropout < 1:
            raise ValueError(f"got {dropout=} but must be in [0, 1)")

        # Split output dimension from the hidden dimensions
        *hidden_dims, out_dim = hidden_dims
        if skip and out_dim != in_dim:
            out_dim = in_dim

        layers = []
        fc_dims = [in_dim, *hidden_dims]
        for d1, d2 in zip(fc_dims[:-1], fc_dims[1:]):
            layers.append(nn.Linear(d1, d2, bias=True))
            if batchnorm:
                layers.append(nn.BatchNorm1d(num_features=d2))
            layers.append(non_linearity())
            if dropout > 0:
                layers.append(nn.Dropout(p=dropout))

        # Always end with FC
        layers.append(nn.Linear(fc_dims[-1], out_dim, bias=True))

        self.fc_layers = nn.Sequential(*layers)
        self.skip = skip

    def forward(self, x):
        x = torch.reshape(x, (x.shape[0], -1))
        z = self.fc_layers(x)

        if self.skip:
            z += x

        return z



from abc import ABC
from typing import Any, Dict, Type, Union, Optional, Sequence, cast

import numpy as np
from numpy import ndarray as Array
from numpy import quantile
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils import check_X_y
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted

from vqr.cvqf import (
    VQF,
    CVQF,
    DiscreteVQF,
    DiscreteCVQF,
    DiscreteVQFBase,
    quantile_levels,
    quantile_contour,
    inversion_sampling,
)
from vqr.solvers import (
    Solver,
    VQESolver,
    VQRSolver,
    CVXVQRSolver,
    POTVQESolver,
    DiscreteSolver,
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)
from vqr.coverage import measure_coverage

SOLVER_TYPES: Dict[str, Type[VQRSolver]] = {
    solver_class.solver_name(): solver_class
    for solver_class in [
        CVXVQRSolver,
        RegularizedDualVQRSolver,
        MLPRegularizedDualVQRSolver,
        POTVQESolver,
    ]
}

DEFAULT_SOLVER_NAME = RegularizedDualVQRSolver.solver_name()


class _Base(BaseEstimator, ABC):
    """
    Base class for vector quantile estimation (VQE) and regression (VQR).
    Compatible with the sklearn API.
    """

    def __init__(
        self,
        solver: Union[str, Solver] = DEFAULT_SOLVER_NAME,
        solver_opts: Optional[Dict[str, Any]] = None,
    ):
        """
        :param solver: Either a supported solver name (see keys of
        :obj:`SOLVER_TYPES`) or an instance of a :class:`VQRSolver`.
        :param solver_opts: If solver is a string, these kwargs will be passed to the
        constructor of the corresponding solver type; If it's an instance,
        it will be copied and these kwargs will be used to override its settings.
        """

        solver_instance: VQRSolver
        solver_opts = solver_opts or {}

        if isinstance(solver, str):
            if solver not in SOLVER_TYPES:
                raise ValueError(
                    f"solver must be one of {[*SOLVER_TYPES.keys()]}, got {solver=}"
                )
            solver_instance = SOLVER_TYPES[solver](**solver_opts)

        elif isinstance(solver, VQRSolver):
            solver_instance = solver.copy(**solver_opts)

        else:
            raise ValueError(
                f"solver must be either a string or an instance of VQRSolver"
            )

        self.solver_opts = solver_opts
        self.solver = solver_instance

        # Deliberate trailing underscore to support detection by check_is_fitted on old
        # versions of sklearn.
        self.fitted_vqf_: Optional[VQF] = None

    def __sklearn_is_fitted__(self):
        return self.fitted_vqf_ is not None

    @property
    def fitted_vqf(self) -> VQF:
        """
        :return: The fitted vector quantile function.
        """
        check_is_fitted(self)
        return self.fitted_vqf_

    @property
    def is_discrete(self) -> bool:
        """
        :return: Whether this estimator solves a discrete problem.
        """
        if self.fitted_vqf_ is None:
            return isinstance(self.solver, DiscreteSolver)
        else:
            return isinstance(self.fitted_vqf, DiscreteVQFBase)

    @property
    def dim_y(self) -> int:
        """
        :return: The dimension of the target variable (Y).
        """
        check_is_fitted(self)
        return self.fitted_vqf.dim_y

    @property
    def quantile_grid(self) -> Sequence[Array]:
        """
        :return: A sequence of quantile level grids as arrays. This is a
            d-dimensional meshgrid (see np.meshgrid) where d is the dimension of the
            target variable Y.
        """
        check_is_fitted(self)
        if not self.is_discrete:
            raise ValueError(f"Quantile grid not defined for non-discrete VQF")
        return cast(DiscreteVQFBase, self.fitted_vqf).quantile_grid

    @property
    def solution_metrics(self) -> Dict[str, Any]:
        """
        :return: A dict containing solver-specific metrics about the solution.
        """
        check_is_fitted(self)
        return self.fitted_vqf.metrics


class VectorQuantileEstimator(_Base):
    """
    Performs vector quantile estimation.
    """

    def __init__(
        self,
        solver: Union[str, VQESolver] = DEFAULT_SOLVER_NAME,
        solver_opts: Optional[Dict[str, Any]] = None,
    ):
        """
        :param solver: Either a supported solver name (see keys of
        :obj:`SOLVER_TYPES`) or an instance of a :class:`VQRSolver`.
        :param solver_opts: If solver is a string, these kwargs will be passed to the
        constructor of the corresponding solver type; If it's an instance,
        it will be copied and these kwargs will be used to override its settings.
        """
        super().__init__(solver, solver_opts)
        assert isinstance(self.solver, VQESolver)

    def vector_quantiles(self, refine: bool = False) -> DiscreteVQF:
        """
        Discretizes the solution and returns all vector quantiles as a DiscreteVQF
        object.

        :param refine: Refine the quantile function using vector monotone rearrangement.
        :return: A DiscreteVQF instance, representing a discretized version of
        the quantile function Q_{Y}(u).
        """

        check_is_fitted(self)
        vqf = self.fitted_vqf
        if isinstance(vqf, DiscreteVQF):
            vqf = vqf.refine() if refine else vqf
        else:
            # TODO: Support non-discrete solvers by discretizing the VQF at some
            #  resolution. Rename to "discretize"?
            raise ValueError(
                f"Currently, obtaining all vector quantiles is only supported for "
                f"discrete solvers."
            )
        return vqf

    def fit(self, Y: Array):
        """
        Fits a quantile estimation model to the given data. In case the data
        is high-dimensional, a vector-quantile model will be fitted.
        :param Y: Data of shape (N, d).
        :return: self.
        """

        # Input validation.
        N = len(Y)
        Y: Array = np.reshape(Y, (N, -1))
        self.fitted_vqf_ = self.solver.solve_vqe(Y=Y)
        return self

    def sample(self, n: int) -> Array:
        """
        Sample from Y based on the fitted vector quantile function Q(u).
        Uses the approach of Inverse transform sampling.
        https://en.wikipedia.org/wiki/Inverse_transform_sampling

        :param n: Number of samples to draw from Y|X=x.
        :return: An array containing the Sampled Y values, of shape (n, d).
        """
        check_is_fitted(self)

        # Calculate vector quantile function
        vqf = self.vector_quantiles()
        q_surfaces = tuple(vqf)  # d x (T, T, ..., T)

        # Sample from the quantile function
        return inversion_sampling(n=n, Qs=q_surfaces)

    def coverage(self, Y: Array, alpha: float = 0.05) -> float:
        """
        Calculates the coverage of given data points using the quantiles fitted by
        this model.

        First, it creates a contour of points in d-dimensional space which surround
        the region in which 1-2*alpha of the distribution (that was corresponds to a
        given quantile function) is contained. Then, the proportion of points
        contained within this contour is calculated.

        :param Y: Points to measure coverage for. Shape should be (N, d).
        :param alpha: Confidence level for the contour.
        :return: The coverage level, between zero and one.
        """
        check_is_fitted(self)

        # Calculate vector quantile functions
        vqf = self.vector_quantiles()
        q_surfaces = tuple(vqf)  # d x (T, T, ..., T)

        return measure_coverage(
            quantile_contour=quantile_contour(Qs=q_surfaces, alpha=alpha)[0],
            data=Y,
        )


class VectorQuantileRegressor(RegressorMixin, _Base):
    """
    Performs vector quantile regression.
    """

    def __init__(
        self,
        solver: Union[str, VQRSolver] = DEFAULT_SOLVER_NAME,
        solver_opts: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(solver, solver_opts)
        assert isinstance(self.solver, VQRSolver)
        self._scaler = StandardScaler(with_mean=True, with_std=True)

    @property
    def fitted_vqf(self) -> CVQF:
        """
        :return: The fitted vector quantile function.
        """
        check_is_fitted(self)
        return self.fitted_vqf_

    @property
    def dim_x(self) -> int:
        """
        :return: The dimension k, of the covariates (X).
        """
        check_is_fitted(self)
        return self.fitted_vqf.dim_x

    def fit(self, X: Array, Y: Array):
        """
        Fits a quantile regression model to the given data. In case the target data
        is high-dimensional, a vector-quantile model will be fitted.
        :param X: Features/covariates of shape (N, k).
        :param Y: Targets/responses of shape (N, d).
        :return: self.
        """

        # Input validation.
        X, y = check_X_y(X, Y, multi_output=True, ensure_2d=True)
        N = len(X)
        Y: Array = np.reshape(Y, (N, -1))

        # Scale features to zero-mean
        X_scaled = self._scaler.fit_transform(X)

        self.fitted_vqf_: CVQF = self.solver.solve_vqr(Y=Y, X=X_scaled)
        return self

    def vector_quantiles(self, X: Array, refine: bool = False) -> Sequence[DiscreteVQF]:
        """
        :param X: Covariates, of shape (N, k).
        :param refine: Whether to refine the conditional quantile function using vector
        monotone rearrangement.
        :return: A sequence of length N, containing DiscreteVQF instances.
        Each element of the sequence corresponds to one of the covariates in X,
        and contains the discretized conditional quantile function Q_{Y|X=x}(u).
        """
        check_is_fitted(self)
        X = self._validate_X_(X, single=False)

        vqf = self.fitted_vqf
        if isinstance(vqf, DiscreteCVQF):
            # Scale X with the fitted transformation before predicting
            X = self._scaler.transform(X)
            vqfs = [vqf.condition(x, refine=refine) for x in X]
        else:
            # TODO: Support non-discrete solvers by discretizing the VQF at some
            #  resolution. Rename to "discretize"?
            raise ValueError(
                f"Currently, obtaining all vector quantiles is only supported for "
                f"discrete solvers."
            )

        return vqfs

    def predict(self, X: Array) -> Array:
        """
        Estimates conditional quantiles of Y|X=x.
        :param X: Samples at which to estimate. Should have shape (N, k).
        :return: An array containing the vector-quantile values.
        Returns the same as vector_quantiles, but stacked into one array.
        The result will be of shape [N, d, T, T, ..., T].
        """
        check_is_fitted(self)

        vqfs: Sequence[DiscreteVQF] = self.vector_quantiles(X)

        # Stack the vector quantiles for each sample into one tensor
        vqs = np.stack(
            # Iterating over vqf produces the quantile surfaces
            [np.stack(vqf, axis=0) for vqf in vqfs],
            axis=0,
        )

        N = X.shape[0] if X is not None else 1
        T = vqfs[0].levels_per_dim
        d = self.dim_y
        assert vqs.shape == (N, d, *[T] * d)
        return vqs

    def sample(self, n: int, x: Array) -> Array:
        """
        Sample from Y|X=x based on the fitted vector quantile function Q(u;x).
        Uses the approach of Inverse transform sampling.
        https://en.wikipedia.org/wiki/Inverse_transform_sampling

        :param n: Number of samples to draw from Y|X=x.
        :param x: One sample of covariates on which to condition Y.
        Should have shape  (k,) or (1, k).
        If None, then samples will be drawn from the unconditional Y.
        :return: An array containing the Sampled Y values, of shape (n, d).
        """
        check_is_fitted(self)
        x = self._validate_X_(X=x, single=True)

        # Calculate vector quantiles given sample X=x
        vqf: DiscreteVQF = self.vector_quantiles(X=x)[0]
        q_surfaces = tuple(vqf)  # d x (T, T, ..., T) where each is d-dimensional

        # Sample from the quantile function
        return inversion_sampling(n=n, Qs=q_surfaces)

    def coverage(self, Y: Array, x: Array, alpha: float = 0.05) -> float:
        """
        Calculates the conditional coverage of given data points using the quantiles
        fitted by this model, conditioned on X=x.

        First, it creates a contour of points in d-dimensional space which surround
        the region in which 100*(1-2*alpha)^d of the distribution (that was corresponds to a
        given quantile function) is contained. Then, the proportion of points
        contained within this contour is calculated.

        :param Y: Points to measure coverage for. Shape should be (N, d).
        :param x: One sample of covariates on which to condition Y.
        Should have shape  (k,) or (1, k).
        :param alpha: Confidence level for the contour.
        :return: The coverage level, between zero and one.
        """
        check_is_fitted(self)
        x = self._validate_X_(X=x, single=True)

        # Calculate vector quantiles given sample X=x
        vqf: DiscreteVQF = self.vector_quantiles(X=x)[0]
        q_surfaces = tuple(vqf)  # d x (T, T, ..., T) where each is d-dimensional

        return measure_coverage(
            quantile_contour=quantile_contour(Qs=q_surfaces, alpha=alpha)[0],
            data=Y,
        )

    def _validate_X_(self, X: Optional[Array], single: bool = False) -> Optional[Array]:
        """
        Validates the shape of a covariates array X.
        :param X: A covariates array, either (n, k) or just (k,). Can be None,
        in which case no validation happens.
        :param single: Whether to validate that n=1.
        :return: X after reshaping to (n,k).
        """
        if X is None:
            raise ValueError("Must provide covariates (X) for VQR")

        error_msg = f"X must be (k,) or ({'1' if single else 'n'}, k), got {X.shape=}"
        if np.ndim(X) == 1 and len(X) == self.dim_x:
            # reshape to (1 ,k)
            X = np.reshape(X, (1, -1))

        elif np.ndim(X) != 2:
            raise ValueError(error_msg)

        if X.shape[1] != self.dim_x:
            raise ValueError(error_msg)

        if single and X.shape[0] != 1:
            # Only a single x is supported by this method
            raise ValueError(error_msg)

        return X


class ScalarQuantileEstimator:
    def __init__(
        self,
        n_levels: int = 50,
    ):
        if n_levels < 2:
            raise ValueError("n_levels must be >= 2")

        self.n_levels = n_levels

    def fit(self, X: Array):
        N = len(X)
        Y = np.reshape(X, (N, -1))
        q = quantile(Y, q=quantile_levels(self.n_levels), axis=0)
        assert q is not None
        assert q.shape[0] == len(quantile_levels(self.n_levels))
        self._alpha = q
        return self

    @property
    def sqr_A(self) -> Array:
        return self._alpha

    @property
    def quantile_levels(self) -> Array:
        """
        :return: An array containing the levels at which the vector quantiles were
            estimated along each target dimension.
        """
        return quantile_levels(self.n_levels)

    @property
    def quantile_values(self) -> Array:
        return self.sqr_A



import os
import re
import logging
from pathlib import Path

import pytest
import nbformat
import nbconvert

from experiments import PROJECT_ROOT_DIR

NOTEBOOKS_DIR = PROJECT_ROOT_DIR.joinpath("notebooks")

TEST_NOTEBOOKS_PATTERN = re.compile(r"\d{2}-[\w-]+\.ipynb")
TEST_NOTEBOOK_PATHS = [
    f for f in NOTEBOOKS_DIR.glob("*") if TEST_NOTEBOOKS_PATTERN.match(f.name)
]
CELL_TIMEOUT_SECONDS = 60 * 5

_LOG = logging.getLogger(__name__)


class TestNotebooks:
    @pytest.fixture(autouse=True)
    def change_cwd(self):
        pwd = os.getcwd()
        try:
            os.chdir(NOTEBOOKS_DIR)
            yield
        finally:
            os.chdir(pwd)

    @pytest.mark.parametrize(
        "notebook_path", TEST_NOTEBOOK_PATHS, ids=[f.stem for f in TEST_NOTEBOOK_PATHS]
    )
    def test_notebook(self, notebook_path: Path):

        _LOG.info(f"Executing notebook {notebook_path}...")

        # Parse notebook
        with open(str(notebook_path), "r") as f:
            nb = nbformat.read(f, as_version=4)

        # Create preprocessor which executes the notebbook in memory - nothing is
        # written back to the file.
        ep = nbconvert.preprocessors.ExecutePreprocessor(
            timeout=CELL_TIMEOUT_SECONDS, kernel_name="python3"
        )

        # Execute. If an exception is raised inside the notebook, this test will fail.
        ep.preprocess(nb)



import logging

import click

from experiments.log import setup_logging
from experiments.base import GPUOptions, OutputOptions, LoggingOptions
from experiments.scale import scale_exp
from experiments.optimization import optim_exp

_LOG = logging.getLogger(__name__)


@click.group(context_settings=dict(show_default=True))
@click.pass_context
@LoggingOptions.cli
@GPUOptions.cli
@OutputOptions.cli
def main(ctx: click.Context, **kw):
    logging_opts = LoggingOptions.parse(ctx)
    setup_logging(level=logging.DEBUG if logging_opts.debug else logging.INFO)


if __name__ == "__main__":
    main.add_command(scale_exp)
    main.add_command(optim_exp)
    main()



import pickle
import logging
from typing import Any, NamedTuple
from hashlib import blake2b

_LOG = logging.getLogger(__name__)


def stable_hash(obj: Any, hash_len: int = 8) -> str:
    """
    :param obj: An object to hash.
    :param hash_len: Desired length of output string.
    :return: A unique and repeatable hash of the given object.
    """
    return blake2b(pickle.dumps(obj), digest_size=hash_len // 2).hexdigest()


def sec_to_time(sec: float):
    """
    Converts a time duration in seconds to days, hours, minutes, seconds and
    milliseconds.
    :param sec: Time in seconds.
    :return: An object with d, h, m, s and ms fields representing the above,
    respectively.
    """

    class __T(NamedTuple):
        d: int
        h: int
        m: int
        s: int
        ms: int

        def __repr__(self):
            return (
                f'{"" if self.d == 0 else f"{self.d}+"}'
                f"{self.h:02d}:{self.m:02d}:{self.s:02d}.{self.ms:03d}"
            )

    if sec < 0:
        raise ValueError("Invalid argument value")

    d = int(sec // (3600 * 24))
    h = int((sec // 3600) % 24)
    m = int((sec // 60) % 60)
    s = int(sec % 60)
    ms = int((sec % 1) * 1000)
    return __T(d, h, m, s, ms)



import logging
from typing import Union, Optional, Sequence

import ot
import numpy as np
import torch
from numpy import ndarray
from torch import Tensor, ones_like
from geomloss import SamplesLoss
from pykeops.torch import Pm, Vi, Vj

from experiments.utils.tensors import ensure_numpy, ensure_torch

_LOG = logging.getLogger(__name__)


def w2_keops(
    Y_gt: Union[ndarray, Tensor],
    Y_est: Union[ndarray, Tensor],
    dtype=torch.float32,
    device: Optional[str] = None,
) -> float:
    """
    Calculates the Wasserstein-2 distance between two distributions, based on their
    samples. Uses pytorch+keops backend.

    :param Y_gt: Ground truth / first distribution's samples.
    :param Y_est: Estimated / second distribution's samples.
    :param dtype: Data type to use for calculation.
    :param device: Device to use for calculation.
    :return: The w2 distance.
    """
    Y_gt = ensure_torch(Y_gt, dtype=dtype, device=device)
    Y_est = ensure_torch(Y_est, dtype=dtype, device=device)

    sample_loss = SamplesLoss(loss="sinkhorn", p=2, blur=0.05)
    w2: Tensor = sample_loss(Y_gt, Y_est)

    return w2.detach().cpu().item()


def w2_pot(
    Y_gt: Union[ndarray, Tensor],
    Y_est: Union[ndarray, Tensor],
    num_iter_max=200_000,
    num_threads=32,
) -> float:
    """
    Calculates the Wasserstein-2 distance between two distributions, based on their
    samples. Uses POT backend with Sinkhorn's algorithm.

    :param Y_gt: Ground truth / first distribution's samples.
    :param Y_est: Estimated / second distribution's samples.
    :param num_iter_max: Maximum iterations for Sinkhorn.
    :param num_threads: Number of threads for solver.
    :return: The w2 distance.
    """

    Y_gt = ensure_numpy(Y_gt)
    Y_est = ensure_numpy(Y_est)

    return ot.emd2(
        a=[],
        b=[],
        M=ot.dist(Y_gt, Y_est),
        numItermax=num_iter_max,
        numThreads=num_threads,
    )


def get_grid_points(Y: Tensor, grid_resolution: int) -> Tensor:
    """
    Get grid points in R^d. The dimension is determined by the dimension of Y.
    Grid borders are determined by the max and min values of Y along each dimension.

    :param Y: Samples in R^d.
    :param grid_resolution: Number of bins to tessellate each dimension.
    :return: Co-ordinates of points in each dimension. Shape=[grid_resolution^d, d]
    """
    Y_max = Y.max(dim=0).values  # torch.tensor([3.0, 2.3]): hack for cond banana plots
    Y_min = Y.min(dim=0).values  # torch.tensor([-3.0, 0.0]): hack for cond banana plots
    axes = [
        torch.linspace(y_i_min.item(), y_i_max.item(), grid_resolution)
        for y_i_min, y_i_max in zip(Y_min, Y_max)
    ]
    grids = torch.meshgrid(axes, indexing="ij")
    grid_points = torch.stack([g.ravel() for g in grids], dim=1)
    return grid_points


def kde_l1(
    Y_gt: Union[ndarray, Tensor],
    Y_est: Union[ndarray, Tensor],
    grid_resolution: int,
    sigma: float,
    dtype: Optional[torch.dtype] = None,
    device: Optional[str] = None,
) -> float:
    """
    Measures L1 distance between the KDEs of two sample distributions.

    :param Y_gt: Samples from the ground-truth distribution
    :param Y_est: Samples from the estimated distribution
    :param grid_resolution: Number of grid points per dimension
    :param sigma: the bandwidth for the kernel in KDE
    :param dtype: Data type to use for calculation.
    :param device: Device on which to execute the metric.
    :return: KDE L1 distance per grid point.
    """
    Y_gt = ensure_torch(Y_gt, dtype=dtype, device=device)
    Y_est = ensure_torch(Y_est, dtype=dtype, device=device)

    assert Y_gt.shape[1] == Y_est.shape[1]  # same dimensions
    grid_shape = [grid_resolution for _ in range(Y_gt.shape[1])]
    grid_points = get_grid_points(Y_gt, grid_resolution)
    kde_gt = kde_keops(Y_gt, grid_points, grid_shape, device, sigma)
    kde_est = kde_keops(Y_est, grid_points, grid_shape, device, sigma)
    return torch.mean(torch.abs(kde_gt - kde_est)).item()


def kde(Y: Tensor, grid_resolution: int, device: str, sigma: float):
    """
    Estimates KDE of the empirical distribution Y.

    :param Y: Samples (empirical distribution).
    :param grid_resolution: Number of grid points per dimension
    :param device: Device on which to execute the metric
    :param sigma: the bandwidth for the kernel in KDE
    :return: KDE L1 distance per grid point.
    """
    return kde_keops(
        Y,
        get_grid_points(Y, grid_resolution),
        [grid_resolution for _ in range(Y.shape[1])],
        device,
        sigma,
    )


def kde_keops(
    Y: Tensor,
    grid: Tensor,
    grid_shape: Sequence[int],
    device: str,
    sigma: float,
):
    """
    Performs kernel density estimation of Y on the grid.

    :param Y: Samples whose KDE needs to be estimated.
    :param grid: Grid points on which the density needs to be estimated.
    :param grid_shape: Shape of the grid.
    :param device: Device to execute on. If gpu needs to be used provide "cuda:{
    device_num}" else, cpu()
    :param sigma: Bandwidth of KDE.
    :return: KDE of the data over the grid.
    """
    assert grid.shape[0] == np.prod(grid_shape)
    assert grid.shape[1] == len(grid_shape)
    d = len(grid_shape)

    dtype = Y.dtype
    Y = Y.contiguous().to(device)
    grid = grid.contiguous().type(dtype).to(device)

    # The distance scales proportional to d, so we need to scale sigma by sqrt(d) to
    # make KDE values computed on different dimensional comparable.
    sigma = Tensor([sigma * np.sqrt(d)]).type(dtype).to(device).contiguous()
    gamma = 1.0 / sigma**2
    b = ones_like(Y[:, 0]).type(dtype).to(device)
    b = b.view(-1, 1).contiguous()
    heatmap = (-Vi(grid).weightedsqdist(Vj(Y), Pm(gamma))).exp() @ b
    heatmap = heatmap.view(*grid_shape).cpu()
    heatmap /= heatmap.sum()
    return heatmap



import logging
import logging.config
from typing import Optional
from pathlib import Path

import yaml

LOGGING_CONFIG = Path(__file__).parent.joinpath("logging.yaml")


def setup_logging(level: Optional[int] = None):
    with open(LOGGING_CONFIG, "r") as f:
        logging_config = yaml.safe_load(f)

    if level is not None:
        for logger_name, logger_settings in logging_config["loggers"].items():
            logger_settings["level"] = logging.getLevelName(level)

    logging.config.dictConfig(logging_config)



import os
import pickle

import numpy as np
import matplotlib.pyplot as plt
from numpy import array
from torch import Tensor
from matplotlib import cm

from experiments.utils.metrics import kde, kde_l1

# CVAE baseline:
# Produced using the open-source implementation by Feldman et al. 2021.
# Available at: https://github.com/shai128/mqr
# Download results from
# https://drive.google.com/drive/folders/1mkrOraclAiPv-YrYibQSBhY_J4MEis9Z?usp=sharing


def plot_kde(kde_map_1, kde_map_2, l1_distance: float, filename: str):
    fig, axes = plt.subplots(1, 2, figsize=(10, 5))
    axes[0].imshow(
        kde_map_1,
        interpolation="bilinear",
        origin="lower",
        cmap=cm.RdPu,
        extent=(0, 1, 0, 1),
    )
    axes[1].imshow(
        kde_map_2,
        interpolation="bilinear",
        origin="lower",
        cmap=cm.RdPu,
        extent=(0, 1, 0, 1),
    )
    plt.title(f"L1 distance: {l1_distance}")
    plt.savefig(f"{filename}.png")


cvae_results_path = "../cvae_results/"
T = 50
GPU_DEVICE_NUM = 0
device = f"cuda:{GPU_DEVICE_NUM}" if GPU_DEVICE_NUM is not None else "cpu"
sigma = 0.1
kde_l1_dists = []

for folder_name in os.listdir(cvae_results_path):
    X = folder_name[-3:]
    with open(f"{cvae_results_path}/x={X}/y_recon_y.pkl", "rb") as f:
        file_contents = pickle.load(f)
        f.close()
    y = Tensor(array(file_contents["y"]))
    y_reconstructed = Tensor(array(file_contents["y_reconstructed"]))

    # Estimate KDEs
    kde_orig = kde(
        y,
        grid_resolution=T * 2,
        device=device,
        sigma=sigma,
    )

    kde_est = kde(
        y_reconstructed,
        grid_resolution=T * 2,
        device=device,
        sigma=sigma,
    )

    # Calculate KDE-L1 distance
    kde_l1_dist = kde_l1(
        y, y_reconstructed, grid_resolution=T * 2, device=device, sigma=sigma
    )

    kde_l1_dists.append(kde_l1_dist)

    # Plot KDEs
    plot_kde(
        kde_orig.T,
        kde_est.T,
        kde_l1_dist,
        f"Y_given_X={float(X):.1f}_cvae",
    )

with open(f"./kde-l1-dists-cvae.pkl", "wb") as f:
    pickle.dump(kde_l1_dists, f)
    print(np.mean(kde_l1_dists))
    f.close()



import random

import numpy as np
import pytest
import matplotlib.pyplot as plt
from sklearn.exceptions import NotFittedError

from vqr import VQRSolver, DiscreteVQF, VectorQuantileRegressor
from vqr.cvqf import DiscreteCVQF
from tests.conftest import _test_monotonicity, monotonicity_offending_projections
from experiments.datasets.mvn import LinearMVNDataProvider
from vqr.solvers.regularized_lse import (
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)


class TestVectorQuantileRegressor(object):
    @pytest.fixture(
        scope="class",
        params=[
            RegularizedDualVQRSolver(
                verbose=True,
                lr=0.5,
                epsilon=1e-2,
            ),
            RegularizedDualVQRSolver(
                verbose=True,
                lr=0.5,
                epsilon=1e-2,
                batchsize_y=1000,
                batchsize_u=100,
            ),
            MLPRegularizedDualVQRSolver(verbose=True, lr=0.5, epsilon=1e-2),
            MLPRegularizedDualVQRSolver(
                verbose=True,
                lr=0.5,
                epsilon=1e-2,
                hidden_layers=[2, 4],
                skip=False,  # No skip, so output will have different k
                num_epochs=1500,
            ),
        ],
        ids=[
            "rvqr_linear",
            "rvqr_linear_batches",
            "rvqr_mlp",
            "rvqr_mlp_change_k",
        ],
    )
    def vqr_solver(self, request, problem_size) -> VQRSolver:
        solver: VQRSolver = request.param
        d, k, N, T = problem_size
        solver = solver.copy(T=T)
        return solver

    @pytest.fixture(
        scope="class",
        params=[
            {"d": 1, "k": 5, "N": 1000, "T": 20},
            {"d": 2, "k": 7, "N": 1000, "T": 20},
            {"d": 3, "k": 3, "N": 1000, "T": 10},
        ],
        ids=[
            "d=1,k=5",
            "d=2,k=7",
            "d=3,k=3",
        ],
    )
    def problem_size(self, request):
        param = request.param
        d, k, N, T = param["d"], param["k"], param["N"], param["T"]
        return (d, k, N, T)

    @pytest.fixture(scope="class")
    def vqr_fitted(self, vqr_solver, problem_size):
        d, k, N, T = problem_size
        X, Y = LinearMVNDataProvider(d=d, k=k).sample(n=N)

        vqr = VectorQuantileRegressor(solver=vqr_solver)
        vqr.fit(X, Y)

        return X, Y, vqr

    def test_shapes(self, vqr_fitted, problem_size):
        d, k, N, T = problem_size
        X, Y, vqr = vqr_fitted

        assert vqr.dim_y == d
        assert vqr.dim_x == k
        assert len(vqr.quantile_grid) == d
        assert all(q.shape == (T,) * d for q in vqr.quantile_grid)
        assert vqr.solution_metrics is not None

        for X_ in [X[[0], :], X[0:, :]]:  # Single and multiple X should be valid
            N_ = len(X_)

            Y_hat = vqr.predict(X_)
            assert Y_hat.shape == (N_, d, *[T] * d)

    def test_vector_quantiles(self, vqr_fitted, problem_size):
        d, k, N, T = problem_size
        X, Y, vqr = vqr_fitted

        # DiscreteVQF per X
        vqfs = vqr.vector_quantiles(X=X)
        assert len(vqfs) == N

        vqf: DiscreteVQF
        for vqf in vqfs:

            # Iterating over DiscreteVQF returns its surfaces
            assert all(vq_surface.shape == (T,) * d for vq_surface in vqf)

            # All values of the quantile function
            assert vqf.values.shape == (d, *[T] * d)
            assert vqf.quantile_grid.shape == (d, *[T] * d)

            for j in range(10):
                # Obtain a random quantile level
                u = np.random.uniform(0.0, 1.0, size=(d,))
                u_idx = np.floor(u * (T - 1)).astype(np.int)

                # Obtain a vector-quantile at level u
                vq = vqf.evaluate(u)

                # d-dimensional vector quantile value
                assert vq.shape == (d,)
                assert np.all(vq == vqf.values[(slice(None), *u_idx)])

    def test_sample(self, vqr_fitted, problem_size, test_out_dir):
        d, k, N, T = problem_size
        X, Y, vqr = vqr_fitted

        n = 1000
        xs = []
        YXs = []
        for i in range(5):
            x = X[i]
            YX_samp = vqr.sample(n, x=x)
            assert YX_samp.shape == (n, d)
            xs.append(x)
            YXs.append(YX_samp)

        if d == 2:
            fig, ax = plt.subplots(1, 1, figsize=(10, 10))
            ax.scatter(Y[:, 0], Y[:, 1], c="k", label="Y (GT)")
            for i, (x, YX_samp) in enumerate(zip(xs, YXs)):
                ax.scatter(
                    YX_samp[:, 0],
                    YX_samp[:, 1],
                    c=f"C{i+1}",
                    label=f"$Y|X=x_{i}$",
                    alpha=0.2,
                )
            ax.legend()
            fig.savefig(test_out_dir.joinpath("vqr_sample.pdf"), bbox_inches="tight")

    @pytest.mark.flaky(reruns=1)
    def test_coverage(self, vqr_fitted, problem_size, test_out_dir):
        d, k, N, T = problem_size
        X, Y, vqr = vqr_fitted

        cov = np.mean(
            [
                # Coverage for each single data point is just 0 or 1
                vqr.coverage(y.reshape(1, d), x=x, alpha=0.05)
                for (x, y) in zip(X, Y)
            ]
        )

        print(f"{cov=}")
        assert cov > (0.6 if d < 3 else 0.2)

    def test_monotonicity(self, vqr_fitted):
        X, Y, vqr = vqr_fitted
        N = len(Y)

        # Test with a few random X's
        idxs = np.random.permutation(N)
        for i in idxs[:10]:
            _test_monotonicity(
                Us=vqr.quantile_grid,
                Qs=list(vqr.vector_quantiles(X=X[[i]], refine=True)[0]),
            )

    @pytest.mark.flaky(reruns=1)
    def test_vector_monotone_rearrangement(self, vqr_solver):
        N = 1000
        d = 2
        k = 3
        T = 15
        X, Y = LinearMVNDataProvider(d=d, k=k, seed=42).sample(n=N)

        vqr_solver = vqr_solver.copy(T=T)
        vqr = VectorQuantileRegressor(solver=vqr_solver)
        vqr.fit(X, Y)

        all_off_projs_refined = []
        all_off_projs = []

        for _ in range(25):
            i = random.randrange(0, N)
            off_projs, all_projs = monotonicity_offending_projections(
                Us=vqr.quantile_grid,
                Qs=list(vqr.vector_quantiles(X=X[[i]], refine=False)[0]),
                projection_tolerance=0.0,
            )

            off_projs_refined, all_projs_refined = monotonicity_offending_projections(
                Us=vqr.quantile_grid,
                Qs=list(vqr.vector_quantiles(X=X[[i]], refine=True)[0]),
                projection_tolerance=0.0,
            )
            all_off_projs.append(len(off_projs) / len(all_projs))
            all_off_projs_refined.append(
                len(off_projs_refined) / len(all_projs_refined)
            )

        print(np.mean(all_off_projs_refined), np.mean(all_off_projs))
        print(np.median(all_off_projs_refined), np.median(all_off_projs))
        assert (np.median(all_off_projs_refined) <= np.median(all_off_projs)) or (
            (np.mean(all_off_projs_refined) <= np.mean(all_off_projs))
        )

    @pytest.mark.parametrize(
        "batch_y", [None, 0.1, 1.0], ids=["y_none", "y_0.1", "y_1.0"]
    )
    @pytest.mark.parametrize(
        "batch_u", [None, 0.2, 1.5], ids=["u_none", "u_0.2", "u_1.5"]
    )
    def test_callback(self, batch_y, batch_u):

        callback_kwargs = []

        def _callback(*args, **kwargs):
            callback_kwargs.append(kwargs)

        N = 1000
        T = 15
        d = 2
        k = 3
        num_epochs = 10

        batchsize_y = int(N * batch_y) if batch_y else None
        batchsize_u = int(T**d * batch_u) if batch_u else None

        num_batches = (
            max(
                np.ceil(N / (batchsize_y or N)),
                np.ceil(T**d / (batchsize_u or T**d)),
            )
            if (batchsize_y or batchsize_u)
            else 1
        )

        X, Y = LinearMVNDataProvider(d=d, k=k).sample(n=N)
        solver = RegularizedDualVQRSolver(
            verbose=False,
            T=T,
            num_epochs=num_epochs,
            batchsize_y=batchsize_y,
            batchsize_u=batchsize_u,
            post_iter_callback=_callback,
        )
        vqr = VectorQuantileRegressor(solver=solver)
        vqr.fit(X, Y)

        assert len(callback_kwargs) == num_epochs * num_batches
        for kw in callback_kwargs:
            assert kw.keys() == callback_kwargs[0].keys()
            solution = kw["solution"]
            assert isinstance(solution, DiscreteCVQF)

    def test_not_fitted(self, vqr_solver):
        X, Y = LinearMVNDataProvider(d=2, k=3).sample(n=100)
        vq = VectorQuantileRegressor(solver=vqr_solver)
        with pytest.raises(NotFittedError):
            _ = vq.quantile_grid
        with pytest.raises(NotFittedError):
            _ = vq.vector_quantiles(X)

    def test_no_x(self, vqr_fitted):
        X, Y, vqr = vqr_fitted
        error_message = "Must provide covariates "
        with pytest.raises(ValueError, match=error_message):
            vqr.sample(n=100, x=None)
        with pytest.raises(ValueError, match=error_message):
            vqr.coverage(Y, x=None)
        with pytest.raises(ValueError, match=error_message):
            vqr.vector_quantiles(X=None)



from typing import Dict, Tuple

import numpy as np
from numpy import array
from numpy.random import shuffle


def split_train_calib_test(
    X: array, Y: array, split_ratios: Tuple[float, float], seed=42
) -> Dict[str, Tuple[array, array]]:
    """
    Randomly splits X and Y into train, calib, and test sets.

    Train set contains `split_ratio[0]*n` points.
    Calibration set contains `(split_ratio[1] - split_ratio[0])*n` points.
    Test set contains `(1-split_ratio[1])*n` points.

    :param X: features sized (n, k)
    :param Y: targets sized (n, d)
    :param split_ratios: Ratios by which to split the dataset (refer to the docstring).
    :param seed: seed the shuffle operation.
    :return: Dictionary with keys ('train', 'calib', 'test') -
    each key consists of X and Y corresponding to train, calibration and test,
    respectively.
    """
    assert X.shape[0] == Y.shape[0]
    N, _ = X.shape
    idxes = np.arange(0, N)
    shuffle(idxes)
    X_train = X[idxes[0 : int(N * split_ratios[0])], :]
    Y_train = Y[idxes[0 : int(N * split_ratios[0])], :]
    X_calib = X[idxes[int(N * split_ratios[0]) : int(N * split_ratios[1])], :]
    Y_calib = Y[idxes[int(N * split_ratios[0]) : int(N * split_ratios[1])], :]
    X_test = X[idxes[int(N * split_ratios[1]) :], :]
    Y_test = Y[idxes[int(N * split_ratios[1]) :], :]
    return {
        "train": (X_train, Y_train),
        "calib": (X_calib, Y_calib),
        "test": (X_test, Y_test),
    }



from abc import ABC, abstractmethod
from typing import Tuple, Optional, Sequence

import numpy as np
from numpy import ndarray as Array


class SerializableRandomDataGenerator(ABC):
    """
    A wrapper for a numpy number generator which allows serialization and
    de-serialization without maintaining the state of the generator.
    This is useful when persistent hashes need to be generated from the provider's
    state.
    """

    def __init__(self, seed: int = 42):
        self._seed = seed
        self._rng = self._create_rng()

    def _create_rng(self) -> np.random.Generator:
        return np.random.default_rng(self._seed)

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_rng"] = None
        return state

    def __setstate__(self, state: dict):
        self.__dict__.update(state)
        self._rng = self._create_rng()


class DataProvider(SerializableRandomDataGenerator):
    """
    Base class for DataProviders which represent a possibly conditional distribution
    of Y|X where Y is d-dimensional and X is k-dimensional. Models the distributions
    of  X, Y|X and the joint (X,Y).
    """

    @property
    @abstractmethod
    def k(self) -> int:
        """
        :return: Number of features (covariates) in X.
        """
        pass

    @property
    @abstractmethod
    def d(self) -> int:
        """
        :return: Number of dimensions in Y.
        """
        pass

    @abstractmethod
    def sample_x(self, n: int) -> Array:
        """
        Samples feature vectors X.
        :param n: Number of samples required.
        :return: An array of shape (n, k) with the samples.
        """
        pass

    @abstractmethod
    def sample(self, n: int, x: Optional[Array] = None) -> Tuple[Array, Array]:
        """
        Samples either (x,y) pairs from the joint distribution or multiple y values
        from Y|X=x.
        :param n: Number of samples to draw.
        :param x: The feature vector to condition on. Should be of shape (1, k).
        If not provided, will be sampled n times.
        :return: A tuple (X, Y) containing n samples from Y|X=x if x was provided (X
        will contain the same sample n times), otherwise n samples from Y where a
        different X was sampled each time, representing samples from the joint
        distribution of (X,Y).
        In all cases, X will be of shape (n, k) and Y will be of shape (n, d).
        """
        pass



from typing import Optional

import ot
import numpy as np
from numpy import ndarray as Array

from vqr.cvqf import DiscreteVQF, DiscreteCVQF, vector_quantile_levels
from vqr.utils import get_kwargs
from vqr.solvers.cvx import SIMILARITY_FN_INNER_PROD
from vqr.solvers.base import VQEDiscreteSolver


class POTVQESolver(VQEDiscreteSolver):
    """
    Solves the VQE problem as a Wasserstein2 (W2) distance between uniform measures
    on U and Y,  with an inner-product ground metric. Uses the POT library's
    implementation of W2.
    """

    def __init__(self, T: int = 50, **emd2_kwargs):
        """
        :param T: Number of quantile levels to estimate along each of the d
        dimensions. The quantile level will be spaced uniformly between 0 and 1.
        :param emd2_kwargs:  Any kwargs supported by pot.emd2().
        """
        self.T = T
        self.emd2_kwargs = emd2_kwargs
        self._solver_opts = get_kwargs()

    @classmethod
    def solver_name(cls) -> str:
        return "vqe_pot"

    @property
    def solver_opts(self) -> dict:
        return self._solver_opts

    @property
    def levels_per_dim(self) -> int:
        return self.T

    def solve_vqe(self, Y: Array) -> DiscreteVQF:
        T = self.T
        N: int = Y.shape[0]
        d: int = Y.shape[1]  # number or target dimensions

        # All quantile levels
        Td: int = T**d

        U: Array = vector_quantile_levels(T, d)
        assert U.shape == (Td, d)

        # Pairwise distances (similarity)
        S: Array = SIMILARITY_FN_INNER_PROD(U, Y.T)  # (Td, d) and (N, d)
        _, log = ot.emd2(
            M=-S,
            a=np.ones([Td]) / Td,
            b=np.ones([N]) / N,
            log=True,
            **self.emd2_kwargs,
        )

        # Obtain the lagrange multipliers Alpha (A) and Beta (B)
        A = np.reshape(-log["u"][:, None], newshape=(Td, 1))
        return DiscreteVQF(T, d, U, A, refine=False)



import pickle

import ot
import numpy as np
import torch
from matplotlib import pyplot as plt

from vqr import VectorQuantileRegressor
from vqr.solvers.cvx import SIMILARITY_FN_INNER_PROD
from vqr.solvers.regularized_lse import (
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)
from experiments.datasets.synthetic_glasses import SyntheticGlassesDataProvider

n = 5000
d = 1
k = 1
T = 100
num_epochs = 40000
linear = False
sigma = 0.1
GPU_DEVICE_NUM = 1
device = f"cuda:{GPU_DEVICE_NUM}" if GPU_DEVICE_NUM is not None else "cpu"
dtype = torch.float32
epsilon = 1e-3
X, Y = SyntheticGlassesDataProvider().sample(n)


if linear:
    solver = RegularizedDualVQRSolver(
        verbose=True,
        num_epochs=num_epochs,
        epsilon=epsilon,
        lr=2.9,
        gpu=True,
        full_precision=False,
        device_num=GPU_DEVICE_NUM,
        lr_factor=0.9,
        lr_patience=500,
        lr_threshold=0.5 * 0.01,
    )
else:
    solver = MLPRegularizedDualVQRSolver(
        verbose=True,
        num_epochs=num_epochs,
        epsilon=epsilon,
        lr=0.4,
        gpu=True,
        skip=True,
        batchnorm=True,
        hidden_layers=(1000, 1000, 1000),
        device_num=GPU_DEVICE_NUM,
        lr_factor=0.9,
        lr_patience=300,
        lr_threshold=0.5 * 0.01,
    )
vqr_est = VectorQuantileRegressor(n_levels=T, solver=solver)

vqr_est.fit(X.reshape(-1, 1), Y.reshape(-1, 1))
levels = vqr_est.quantile_levels
quantiles = []
for i in range(n):
    X_i = np.array([X[i]])[:, None]
    quantile_est = vqr_est.vector_quantiles(X=X_i)[0][0]
    cost = -SIMILARITY_FN_INNER_PROD(levels[:, None], quantile_est[None, :])
    pi = ot.emd([], [], cost)
    refined_quantiles = T * pi @ quantile_est
    quantiles.append(refined_quantiles)

quantiles = np.stack(quantiles, axis=0)
Y_est = np.concatenate(
    [vqr_est.sample(n=1, x=np.array([X[i]])[:, None])[0] for i in range(n)]
)

fig, ax = plt.subplots(1, 2, figsize=(10, 5))
ax[0].plot(X, Y, ".", label="GT Y|X")
ax[0].set_ylabel("y")
ax[0].set_xlabel("x")
ax[0].set_title("GT")
ax[0].legend()
ax[1].plot(X, Y_est, ".", label="samples of Y|X")
for quantile_level in np.linspace(0.1, 0.90, 10):
    ax[1].plot(X, quantiles[:, int(quantile_level * T)], label=f"{quantile_level:.2f}")
ax[1].legend()
ax[1].set_ylabel("y")
ax[1].set_xlabel("x")
ax[1].set_title("Non-linear QR")
plt.suptitle("Simultaneous scalar QR", fontsize="x-large")
plt.tight_layout()
plt.savefig(f"quantile_levels_{linear=}.png")

with open(f"quantiles_{linear=}.pkl", "wb") as f:
    pickle.dump({"quantiles": quantiles}, f)



import numpy as np
import pytest
from matplotlib import pyplot as plt
from sklearn.exceptions import NotFittedError

from vqr import DiscreteVQF, VectorQuantileEstimator
from tests.conftest import _test_monotonicity, monotonicity_offending_projections
from vqr.solvers.cvx import CVXVQRSolver
from vqr.solvers.pot import POTVQESolver
from experiments.datasets.mvn import IndependentDataProvider


class TestVectorQuantileEstimator(object):
    @pytest.fixture(
        scope="class",
        params=[
            (
                "regularized_dual",
                {"verbose": True, "lr": 0.1, "epsilon": 1e-6},
            ),
            ("cvx_primal", {"verbose": True}),
            ("vqe_pot", {}),
        ],
        ids=["rvqr_linear", "cvx_primal", "vqe_pot"],
    )
    def solver(self, request):
        return request.param

    @pytest.fixture(
        scope="class",
        params=[
            {"d": 1, "N": 200, "T": 20},
            {"d": 2, "N": 200, "T": 20},
            {"d": 3, "N": 200, "T": 10},
        ],
        ids=[
            "d=1",
            "d=2",
            "d=3",
        ],
    )
    def problem_size(self, request):
        param = request.param
        d, N, T = param["d"], param["N"], param["T"]
        return (d, N, T)

    @pytest.fixture(scope="class")
    def vqe_fitted(self, solver, problem_size):
        d, N, T = problem_size
        solver_name, solver_opts = solver
        solver_opts["T"] = T

        X, Y = IndependentDataProvider(d=d, k=1).sample(n=N)
        vqe = VectorQuantileEstimator(
            solver=solver_name,
            solver_opts=solver_opts,
        )
        vqe.fit(Y)

        return Y, vqe

    def test_shapes(self, vqe_fitted, problem_size):
        d, N, T = problem_size
        Y, vqe = vqe_fitted

        assert vqe.dim_y == d
        assert len(vqe.quantile_grid) == d
        assert all(q.shape == (T,) * d for q in vqe.quantile_grid)

    def test_vector_quantiles(self, vqe_fitted, problem_size):
        d, N, T = problem_size
        Y, vqe = vqe_fitted

        vqf: DiscreteVQF = vqe.vector_quantiles()
        assert len(vqf) == d
        assert all(q_surface.shape == (T,) * d for q_surface in vqf)
        assert vqf.values.shape == (d, *[T] * d)
        assert vqf.quantile_grid.shape == (d, *[T] * d)

    def test_sample(self, vqe_fitted, problem_size, test_out_dir):
        d, N, T = problem_size
        Y, vqe = vqe_fitted

        n = 1000
        Y_samp = vqe.sample(n)
        assert Y_samp.shape == (n, d)

        if d == 2:
            fig, ax = plt.subplots(1, 1, figsize=(10, 10))
            ax.scatter(Y[:, 0], Y[:, 1], c="k", label="Y (GT)")
            ax.scatter(Y_samp[:, 0], Y_samp[:, 1], c="C0", label="Y")
            ax.legend()
            fig.savefig(test_out_dir.joinpath("vqe_sample.pdf"), bbox_inches="tight")

    @pytest.mark.flaky(reruns=1)
    def test_coverage(self, vqe_fitted, problem_size, test_out_dir):
        d, N, T = problem_size
        Y, vqe = vqe_fitted

        cov = vqe.coverage(Y, alpha=0.05)
        print(f"{cov=}")
        assert cov > (0.7 if d < 3 else 0.25)

    def test_not_fitted(self, solver):
        solver_name, solver_opts = solver
        vq = VectorQuantileEstimator(solver=solver_name, solver_opts=solver_opts)
        with pytest.raises(NotFittedError):
            _ = vq.quantile_grid
        with pytest.raises(NotFittedError):
            _ = vq.vector_quantiles()

    def test_monotonicity(self, vqe_fitted):
        Y, vqe = vqe_fitted

        _test_monotonicity(
            Us=vqe.quantile_grid,
            Qs=list(vqe.vector_quantiles(refine=True)),
        )

    def test_vmr_vqe(self, solver):
        solver_name, solver_opts = solver

        T = 15
        N = 1000
        d = 2
        solver_opts["T"] = T

        _, Y = IndependentDataProvider(d=d, k=1).sample(n=N)
        vqe = VectorQuantileEstimator(solver=solver_name, solver_opts=solver_opts)
        vqe.fit(Y)

        off_projs, _ = monotonicity_offending_projections(
            Us=vqe.quantile_grid,
            Qs=list(vqe.vector_quantiles(refine=False)),
            projection_tolerance=0.0,
        )

        off_projs_refined, _ = monotonicity_offending_projections(
            Us=vqe.quantile_grid,
            Qs=list(vqe.vector_quantiles(refine=True)),
            projection_tolerance=0.0,
        )

        print(len(off_projs), len(off_projs_refined))
        assert len(off_projs_refined) <= len(off_projs)


class TestPOTVQE:
    # Important that N is odd
    @pytest.mark.parametrize("N", [101, 201, 301, 1001])
    @pytest.mark.parametrize("T", [10, 20])
    def test_pot_quantile_vs_sorted_Y_1d(self, N, T):
        Y = np.random.randn(N, 1)
        est_quantiles = POTVQESolver(T=T).solve_vqe(Y=Y).values[0]
        sorted_Y = np.sort(Y.squeeze())[:: int(N / T)]
        assert np.allclose(est_quantiles[1:], sorted_Y[1:-1])

    @pytest.mark.parametrize("N", [101, 201, 301])
    @pytest.mark.parametrize("T", [5, 10])
    @pytest.mark.parametrize("d", [1, 2, 3])
    def test_pot_quantile_vs_sorted_Y_2d(self, N, T, d):
        Y = np.random.randn(N, d)
        est_quantiles = POTVQESolver(T=T).solve_vqe(Y=Y).values
        est_quantiles_cvx = CVXVQRSolver(T=T).solve_vqe(Y=Y).values
        for Q, Q_cvx in zip(est_quantiles, est_quantiles_cvx):
            assert Q.shape == (T,) * d
            assert Q_cvx.shape == (T,) * d
            assert np.allclose(Q, Q_cvx)

    @pytest.mark.parametrize("T", [10, 20, 50, 100])
    @pytest.mark.parametrize("N", [101, 201, 301, 1001])
    def test_vqe_vs_cvx(self, T, N):
        Y = np.random.randn(N, 1)
        est_quantiles = POTVQESolver(T=T).solve_vqe(Y=Y).values[0]
        est_quantiles_cvx = CVXVQRSolver(T=T).solve_vqe(Y=Y).values[0]
        assert np.allclose(est_quantiles, est_quantiles_cvx)

    @pytest.mark.parametrize("d", [1, 2, 3, 4])
    def test_multiple_d(self, d):
        T = 10
        Y = np.random.randn(1000, d)
        quantiles = POTVQESolver(T=T).solve_vqe(Y=Y).values
        assert len(quantiles) == d

    @pytest.mark.parametrize("N", [1001, 2001, 3001])
    @pytest.mark.parametrize("T", [5, 10, 20, 25, 30])
    def test_comonotonicity(self, N, T):
        Y = np.random.randn(N, 2)
        vqf = POTVQESolver(T=T).solve_vqe(Y=Y)
        _test_monotonicity(Us=vqf.quantile_grid, Qs=vqf.values)



import torch
import pytest

from vqr.models import MLP


class TestMLP(object):
    @pytest.fixture(autouse=True)
    def setup(self):
        pass

    @pytest.mark.parametrize(
        "in_dim, hidden_dims, nl, skip, batchnorm, dropout",
        [
            [3, [5], "relu", False, True, 0.1],
            [3, [5, 10], "relu", False, True, 0.1],
            [3, [5, 10, 3], "relu", True, True, 0.1],
            [3, [5, 10, 5], "relu", True, True, 0.1],
            [3, "5, 10,5", "relu", True, True, 0.1],
            #
            [3, [5], "relu", False, True, 0],
            [3, [5, 10], "relu", False, True, 0],
            [3, [5, 10, 3], "relu", True, True, 0],
            [3, [5, 10, 5], "relu", True, True, 0],
            [3, ",5,10, 5 ", "relu", True, True, 0],
            #
            [3, [5], "tanh", False, False, 0.1],
            [3, [5, 10], "tanh", False, False, 0.1],
            [3, [5, 10, 3], "tanh", True, False, 0.2],
            [3, [5, 10, 5], "tanh", True, False, 0.2],
            #
            [3, [5], "tanh", False, False, 0],
            [3, [5, 10], "tanh", False, False, 0],
            [3, [5, 10, 3], "tanh", True, False, 0],
            [3, [5, 10, 5], "tanh", True, False, 0],
            [3, "5,10,5", "tanh", True, False, 0],
        ],
    )
    def test_mlp(self, in_dim, hidden_dims, nl, skip, batchnorm, dropout):
        mlp = MLP(
            in_dim=in_dim,
            hidden_dims=hidden_dims,
            nl=nl,
            skip=skip,
            batchnorm=batchnorm,
            dropout=dropout,
        )

        N = 100
        x = torch.randn(N, in_dim)
        y = mlp(x)

        print(mlp)
        assert isinstance(mlp.fc_layers[-1], torch.nn.Linear)
        if skip:
            assert y.shape == (N, in_dim)
        else:
            assert y.shape == (N, hidden_dims[-1])

    @pytest.mark.parametrize(
        "in_dim, hidden_dims, nl, skip, batchnorm, dropout, error_msg",
        [
            # no hidden dims
            [3, [], "relu", False, True, 0.1, "hidden_dims="],
            [3, None, "relu", False, True, 0.1, "hidden_dims="],
            # invalid hidden dims
            [3, "", "relu", False, True, 0.1, "hidden_dims is a string"],
            [3, ",", "relu", False, True, 0.1, "hidden_dims is a string"],
            [3, "1,2,a,3", "relu", False, True, 0.1, "hidden_dims is a string"],
            # unknown nl
            [3, [5], "wtf", False, True, 0.1, "nl="],
            # invalid dropout
            [3, [5], "relu", False, True, -0.1, "dropout="],
            [3, [5], "relu", False, True, 1.0, "dropout="],
            [3, [5], "relu", False, True, 1.1, "dropout="],
        ],
    )
    def test_init_exceptions(
        self, in_dim, hidden_dims, nl, skip, batchnorm, dropout, error_msg
    ):
        with pytest.raises(ValueError, match=error_msg):
            MLP(
                in_dim=in_dim,
                hidden_dims=hidden_dims,
                nl=nl,
                skip=skip,
                batchnorm=batchnorm,
                dropout=dropout,
            )
